from . import (
    annotations,
    area,
    arrow,
    calculus,
    circle,
    clip,
    coordinates,
    CTM,
    definition,
    diagram,
    diffeqs,
    graph,
    grid_axes,
    group,
    implicit,
    label,
    line,
    math_utilities,
    network,
    parametric_curve,
    parse,
    path,
    point,
    polygon,
    rectangle,
    repeat,
    riemann_sum,
    shape,
    slope_field,
    tags,
    tangent_line,
    user_namespace,
    utilities,
    vector
)


